'use strict';

var lpm = require('length-prefixed-message');
var PROTOCOL_ID = require('./protocol-id');
var varint = require('varint');
var range = require('lodash.range');
var parallel = require('run-parallel');

exports = module.exports = Multistream;

function Multistream(conn, isListener, readyCallback) {
  var _this = this;

  if (!(this instanceof Multistream)) {
    return new Multistream(conn, isListener, readyCallback);
  }

  if (typeof isListener === 'function') {
    readyCallback = isListener;
    isListener = false;
  }

  if (!readyCallback) {
    readyCallback = function noop() {};
  }

  var handlers = {};

  this.addHandler = function (protocol, handlerFunc) {
    if (!isListener) {
      throw new Error('This conn is not on listener mode, can not handle');
    }

    if (typeof handlerFunc !== 'function') {
      throw new Error('handler function must be a function');
    }

    handlers[protocol] = handlerFunc;
  };

  this.select = function (protocol, callback) {
    if (isListener) {
      return callback(new Error('This conn is on listener mode, can not select'));
    }
    lpm.write(conn, new Buffer(protocol + '\n'));
    lpm.read(conn, function (msgBuffer) {
      var msg = msgBuffer.toString().slice(0, -1);
      if (msg === protocol) {
        return callback(null, conn);
      }
      if (msg === 'na') {
        return callback(new Error(protocol + ' not supported'));
      }
    });
  };

  this.ls = function (callback) {
    if (isListener) {
      return callback(new Error('This conn is on listener mode, can not ls'));
    }

    lpm.write(conn, new Buffer('ls' + '\n'));
    var protos = [];
    lpm.read(conn, function (msgBuffer) {
      var size = varint.decode(msgBuffer); // eslint-disable-line
      var nProtos = varint.decode(msgBuffer, varint.decode.bytes);

      times(nProtos, function (n, next) {
        lpm.read(conn, function (msgBuffer) {
          protos.push(msgBuffer.toString().slice(0, -1));
          next();
        });
      }, function (err) {
        if (err) {
          return callback(err);
        }
        callback(null, protos);
      });
    });
  };

  if (isListener) {
    lpm.write(conn, new Buffer(PROTOCOL_ID + '\n'));
    listenerMultistreamHandshakeCheck();
  } else {
    lpm.read(conn, function (buffer) {
      var msg = buffer.toString().slice(0, -1);
      if (msg === PROTOCOL_ID) {
        lpm.write(conn, new Buffer(PROTOCOL_ID + '\n'));
        readyCallback(null, _this);
      } else {
        readyCallback(new Error('Incompatible multistream'));
      }
    });
  }

  function listenerMultistreamHandshakeCheck() {
    var _this2 = this;

    lpm.read(conn, function (buffer) {
      var msg = buffer.toString().slice(0, -1);
      if (msg === PROTOCOL_ID) {
        lpm.read(conn, incMsg);
        readyCallback(null, _this2);
      } else {
        // TODO This would be where we try to support other versions
        // of multistream (backwards compatible). Currently we have
        // just one, so this never happens.
        return readyCallback(new Error('not supported version of multistream'));
      }
    });
  }

  function incMsg(msgBuffer) {
    var msg = msgBuffer.toString().slice(0, -1);

    if (msg === 'ls') {
      var protos = Object.keys(handlers);
      var nProtos = protos.length;
      // total size of the list of protocols, including varint and newline
      var size = protos.reduce(function (size, proto) {
        var p = new Buffer(proto + '\n');
        var el = varint.encodingLength(p.length);
        return size + el;
      }, 0);

      var nProtoVI = new Buffer(varint.encode(nProtos));
      var sizeVI = new Buffer(varint.encode(size));
      var buf = Buffer.concat([nProtoVI, sizeVI, new Buffer('\n')]);
      lpm.write(conn, buf);
      protos.forEach(function (proto) {
        lpm.write(conn, new Buffer(proto + '\n'));
      });
    }

    if (handlers[msg]) {
      // Protocol supported, ACK back
      lpm.write(conn, new Buffer(msg + '\n'));
      return handlers[msg](conn);
    } else {
      // Protocol not supported, wait for new handshake
      lpm.write(conn, new Buffer('na' + '\n'));
    }

    // continue listening
    lpm.read(conn, incMsg);
  }
}

function times(i, work, callback) {
  parallel(range(i).map(function (i) {
    return function (callback) {
      return work(i, callback);
    };
  }), callback);
}